package fe

import (
	"log"

	"git.autistici.org/ale/autoradio"
	"github.com/jmcvetta/randutil"
)

// A load balancing policy selects a single node from the pool of
// currently active ones.
type LoadBalancingPolicy interface {
	GetNode([]*autoradio.NodeStatus) *autoradio.NodeStatus
}

// Simple load balancing policy that always returns the nodes with the
// least amount of listeners.
type leastListenersPolicy struct{}

func (llp leastListenersPolicy) GetNode(nodes []*autoradio.NodeStatus) *autoradio.NodeStatus {
	var minNode *autoradio.NodeStatus
	min := 1000000
	for _, n := range nodes {
		if listeners := n.NumListeners(); listeners < min {
			minNode = n
			min = listeners
		}
	}
	return minNode
}

// Simple load balancing policy that selects a node randomly,
// associating a weight to each node inversely proportional to the
// number of active listeners.
type weightedListenersPolicy struct{}

func (wlp weightedListenersPolicy) GetNode(nodes []*autoradio.NodeStatus) *autoradio.NodeStatus {
	choices := make([]randutil.Choice, 0, len(nodes))
	weightBase := 1000000
	for _, n := range nodes {
		w := weightBase / (n.NumListeners() + 1)
		choices = append(choices, randutil.Choice{w, n})
	}
	result, err := randutil.WeightedChoice(choices)
	if err != nil {
		return nil
	}
	return result.Item.(*autoradio.NodeStatus)
}

// Load balancing policy that will send requests to the backend whose
// bandwidth utilization is the lowest.
type leastBandwidthPolicy struct{}

func (lbp leastBandwidthPolicy) GetNode(nodes []*autoradio.NodeStatus) *autoradio.NodeStatus {
	// Capping usage at 100%, nodes above that won't be returned
	// at all.
	var minNode *autoradio.NodeStatus
	minUtil := float64(1.0)
	for _, n := range nodes {
		if n.BandwidthUsage < minUtil {
			minUtil = n.BandwidthUsage
			minNode = n
		}
	}
	return minNode
}

func getNamedLoadBalancingPolicy(lbpolicy string) LoadBalancingPolicy {
	switch lbpolicy {
	case "leastloaded", "ll":
		return &leastListenersPolicy{}
	case "weighted", "wl":
		return &weightedListenersPolicy{}
	case "leastbandwidth", "lb":
		return &leastBandwidthPolicy{}
	}
	log.Fatalf("Unknown load-balancing policy '%s'", lbpolicy)
	return nil
}
